@@ -14,15 +14,18 @@ import mimetypes
 import subprocess       # nosec
 import pandas as pd
 import tornado.template
+from lxml import etree
 from threading import Thread
 from six.moves.queue import Queue
 from orderedattrdict import AttrDict
 from tornado.concurrent import Future
 from tornado.ioloop import IOLoop, PeriodicCallback
 from gramex.config import app_log, merge, used_kwargs, CustomJSONDecoder, CustomJSONEncoder
+from gramex.config import PathConfig
 from six.moves.urllib_parse import urlparse
 
 
+_undef = object()
 MILLISECOND = 0.001         # in seconds
 _opener_defaults = dict(mode='r', buffering=-1, encoding='utf-8', errors='strict',
                         newline=None, closefd=True)
@@ -126,6 +129,65 @@ def _template(path, **kwargs):
     return tornado.template.Loader(root, **kwargs).load(name)
 
 
+def read_excel(io, sheet_name=0, table=None, name=None, range=None, header=_undef, **kwargs):
+    '''
+    Read data from an Excel file as a DataFrame
+
+    :arg str/file io: path or file-like object pointing to an Excel file
+    :arg str/int sheet_name: sheet to load data from. Sheet names are specified as strings.
+        Integers pick zero-indexed sheet position. default: 0
+    :arg str table: Worksheet table to load from sheet, e.g. ``'Table1'``
+    :arg str name: Defined name to load from sheet, e.g. ``'MyNamedRange'``
+    :arg str range: Cell range to load from sheet, e.g. ``'A1:C10'``
+    :arg None/int/list[int] header: Row (0-indexed) to use for the column labels.
+        A list of integers is combined into a MultiIndex. Use None if there is no header.
+
+    If none of ``table``, ``name``, ``range`` are specified, this loads the entire sheet using
+    ``pd.read_excel``. All other keyword arguments are passed through to ``pd.read_excel``.
+
+    If any of these are specified, we use ``openpyxl`` to read a specific cell range and infer
+    column types. ``table`` overrides ``name`` overrides ``range``.
+    '''
+    if not any((range, name, table)):
+        return pd.read_excel(io, sheet_name=sheet_name, header=0 if header is _undef else header,
+                             **kwargs)
+
+    import openpyxl
+    wb = openpyxl.load_workbook(io, data_only=True)
+    # Pick a SINGLE sheet using sheet_name -- it can be an int or a str
+    ws = wb[wb.sheetnames[sheet_name] if isinstance(sheet_name, int) else sheet_name]
+    # Get the data range to be picked
+    if table is not None:
+        range = ws.tables[table].ref
+        # Tables themselves specify whether they have a column header. Use this as default
+        if header is _undef:
+            header = list(__builtins__['range'](ws.tables[table].headerRowCount))
+    elif name is not None:
+        # If the name is workbook-scoped, get it directly
+        defined_name = wb.defined_names.get(name)
+        # Else, if it's sheet-scoped, get it related to the sheet
+        if defined_name is None:
+            defined_name = wb.defined_names.get(name, wb.sheetnames.index(ws.title))
+        # Raise an error if we can't find it
+        if defined_name is None:
+            raise ValueError(f'{io}: missing name {name} in sheet {sheet_name}')
+        # Note: This only works if it's a cell range. If we create a named range inside a table,
+        # Excel may store this as =Table[[#All],[Col1]:[Col5]], which isn't a valid range.
+        # Currently, we ignore that, and assumed that the name is like Sheet1!A1:C10
+        range = defined_name.attr_text.split('!')[-1]
+
+    data = pd.DataFrame([[cell.value for cell in row] for row in ws[range]])
+    # Header defaults to 0 if undefined. If it's not None, apply the header
+    header = 0 if header is _undef else header
+    if header is not None:
+        data = (data.T.set_index(header).T      # Set header rows as column names
+                    .reset_index(drop=True)     # Drop index with "holes" where headers were
+                    .rename_axis(               # Column name has header index (e.g. 0). Drop it
+                        [None] * len(header) if isinstance(header, (list, tuple)) else None,
+                        axis=1))
+    return data.infer_objects()                 # Convert data types
+
+
 def stat(path):
     '''
     Returns a file status tuple - based on file last modified time and file size
@@ -151,17 +213,18 @@ def hashed(val):
 # gramex.cache.open() stores its cache here.
 # {(path, callback): {data: ..., stat: ...}}
 _OPEN_CACHE = {}
-_OPEN_CALLBACKS = dict(
+open_callback = dict(
     bin=opener(None, read=True, mode='rb', encoding=None, errors=None),
     txt=opener(None, read=True),
     text=opener(None, read=True),
     csv=pd.read_csv,
-    excel=pd.read_excel,
-    xls=pd.read_excel,
-    xlsx=pd.read_excel,
+    excel=read_excel,
+    xls=read_excel,
+    xlsx=read_excel,
     hdf=pd.read_hdf,
     h5=pd.read_hdf,
     html=pd.read_html,
+    json=opener(json.load),
     jsondata=pd.read_json,
     sas=pd.read_sas,
     stata=pd.read_stata,
@@ -172,6 +235,11 @@ _OPEN_CALLBACKS = dict(
     markdown=_markdown,
     tmpl=_template,
     template=_template,
+    xml=etree.parse,
+    svg=etree.parse,
+    rss=etree.parse,
+    atom=etree.parse,
+    config=PathConfig,
     yml=_yaml,
     yaml=_yaml
 )
@@ -198,7 +266,7 @@ def open(path, callback=None, transform=None, rel=False, **kwargs):
     - ``jsondata``: reads files using pd.read_json
     - ``template``: reads files using tornado.Template via io.open
     - ``markdown`` or ``md``: reads files using markdown.markdown via io.open
-    - ``csv``, ``excel``, ``xls``, `xlsx``, ``hdf``, ``h5``, ``html``, ``sas``,
+    - ``csv``, ``excel``, ``xls``, ``xlsx``, ``hdf``, ``h5``, ``html``, ``sas``,
       ``stata``, ``table``, ``parquet``, ``feather``: reads using Pandas
     - ``xml``, ``svg``, ``rss``, ``atom``: reads using lxml.etree
 
@@ -221,6 +289,12 @@ def open(path, callback=None, transform=None, rel=False, **kwargs):
     This is called as ``my_format_reader_function('data.fmt', arg='value')`` and
     cached. Future calls do not re-load and re-calculate this data.
 
+    To support a new callback string, set ``gramex.cache.open_callback[key] = method``.
+    For example::
+
+        gramex.cache.open_callback['shp'] = geopandas.read_file     # Register
+        prs = gramex.cache.open('my.shp', layer='countries')        # Open with method
+
     ``transform=`` is an optional function that processes the data returned by
     the callback. For example::
 
@@ -270,18 +344,8 @@ def open(path, callback=None, transform=None, rel=False, **kwargs):
             data = callback(path, **kwargs)
         elif callback_is_str:
             method = None
-            if callback in _OPEN_CALLBACKS:
-                method = _OPEN_CALLBACKS[callback]
-            elif callback in {'json'}:
-                import json
-                method = opener(json.load)
-            elif callback in {'config'}:
-                from gramex.config import PathConfig
-                method = PathConfig
-            elif callback in {'xml', 'svg', 'rss', 'atom'}:
-                from lxml import etree
-                method = etree.parse
-
+            if callback in open_callback:
+                method = open_callback[callback]
             if method is not None:
                 data = method(path, **kwargs)
             elif original_callback is None:
@@ -429,9 +493,6 @@ def query(sql, engine, state=None, **kwargs):
     _cache = kwargs.pop('_cache', _QUERY_CACHE)
     store_cache = True
 
-    key = (str(sql), json.dumps(kwargs.get('params', {}), sort_keys=True), engine.url)
-    cached = _cache.get(key, {})
-    current_status = cached.get('status', None) if cached else None
     if isinstance(state, (list, tuple)):
         status = _table_status(engine, tuple(state))
     elif isinstance(state, six.string_types):
@@ -446,7 +507,8 @@ def query(sql, engine, state=None, **kwargs):
         raise TypeError('gramex.cache.query(state=) must be a table list, query or fn, not %s',
                         repr(state))
 
-    if status == current_status:
+    key = (str(sql), json.dumps(kwargs.get('params', {}), sort_keys=True), engine.url)
+    if key in _cache and _cache[key]['status'] == status:
         result = _cache[key]['data']
     else:
         app_log.debug('gramex.cache.query: %s. engine: %s. state: %s. kwargs: %s', sql, engine,
