@@ -7,11 +7,10 @@ import re
 import six
 import time
 import json
-import sqlalchemy
+import sqlalchemy as sa
 import pandas as pd
 import gramex.cache
 from tornado.escape import json_encode
-from sqlalchemy.sql import text
 from gramex.config import merge, app_log
 from orderedattrdict import AttrDict
 
@@ -67,7 +66,7 @@ def _replace(engine, args, *vars, **kwargs):
     return _format(list(vars)) + [_format(kwargs)]
 
 
-def filter(url, args={}, meta={}, engine=None, table=None, ext=None,
+def filter(url, args={}, meta={}, engine=None, ext=None, columns=None,
            query=None, queryfile=None, transform=None, transform_kwargs=None, **kwargs):
     '''
     Filters data using URL query parameters. Typical usage::
@@ -84,9 +83,14 @@ def filter(url, args={}, meta={}, engine=None, table=None, ext=None,
     :arg dict meta: this dict is updated with metadata during the course of filtering
     :arg str engine: over-rides the auto-detected engine. Can be 'dataframe', 'file',
         'http', 'https', 'sqlalchemy', 'dir'
-    :arg str table: table name (if url is an SQLAlchemy URL), ``.format``-ed
-        using ``args``.
     :arg str ext: file extension (if url is a file). Defaults to url extension
+    :arg dict columns: database column names to create if required (if url is a database).
+        Keys are column names. Values can be SQL types, or dicts with these keys:
+            - ``type`` (str), e.g. ``"VARCHAR(10)"``
+            - ``default`` (str/int/float/bool), e.g. ``"none@example.org"``
+            - ``nullable`` (bool), e.g. ``False``
+            - ``primary_key`` (bool), e.g. ``True`` -- used only when creating new tables
+            - ``autoincrement`` (bool), e.g. ``True`` -- used only when creating new tables
     :arg str query: optional SQL query to execute (if url is a database),
         ``.format``-ed using ``args`` and supports SQLAlchemy SQL parameters.
         Loads entire result in memory before filtering.
@@ -102,7 +106,10 @@ def filter(url, args={}, meta={}, engine=None, table=None, ext=None,
     :return: a filtered DataFrame
 
     Remaining kwargs are passed to :py:func:`gramex.cache.open` if ``url`` is a file, or
-    ``sqlalchemy.create_engine`` if ``url`` is a SQLAlchemy URL.
+    ``sqlalchemy.create_engine`` if ``url`` is a SQLAlchemy URL. In particular:
+
+    :arg str table: table name (if url is an SQLAlchemy URL), ``.format``-ed
+        using ``args``.
 
     If this is used in a handler as::
 
@@ -182,6 +189,7 @@ def filter(url, args={}, meta={}, engine=None, table=None, ext=None,
     - ``limit``: Limit as integer - ``None`` if limit is not applied
     - ``count``: Total number of rows, if available
     - ``by``: Group by columns as ``[col, ...]``
+    - ``inserted``: List of (dict of primary values) for each inserted row
 
     These variables may be useful to show additional information about the
     filtered data.
@@ -201,8 +209,8 @@ def filter(url, args={}, meta={}, engine=None, table=None, ext=None,
     })
     controls = _pop_controls(args)
     transform = _transform_fn(transform, transform_kwargs)
-    url, table, ext, query, queryfile, kwargs = _replace(
-        engine, args, url, table, ext, query, queryfile, **kwargs)
+    url, ext, query, queryfile, kwargs = _replace(
+        engine, args, url, ext, query, queryfile, **kwargs)
 
     # Use the appropriate filter function based on the engine
     if engine == 'dataframe':
@@ -219,19 +227,21 @@ def filter(url, args={}, meta={}, engine=None, table=None, ext=None,
         data = gramex.cache.open(url, ext, transform=transform, **kwargs)
         return _filter_frame(data, meta=meta, controls=controls, args=args)
     elif engine == 'sqlalchemy':
-        engine = create_engine(url, **kwargs)
+        table = kwargs.pop('table', None)
+        state = kwargs.pop('state', None)
+        engine = alter(url, table, columns, **kwargs)
         if query or queryfile:
             if queryfile:
                 query = gramex.cache.open(queryfile, 'text')
-            state = None
-            if isinstance(table, six.string_types):
-                state = table if ' ' in table else [table]
-            elif isinstance(table, (list, tuple)):
-                state = [t for t in table]
-            elif table is not None:
-                raise ValueError('table: must be string or list of strings, not %r' % table)
+            if not state:
+                if isinstance(table, six.string_types):
+                    state = table if ' ' in table else [table]
+                elif isinstance(table, (list, tuple)):
+                    state = [t for t in table]
+                elif table is not None:
+                    raise ValueError('table: must be string or list of strings, not %r' % table)
             all_params = {k: v[0] for k, v in args.items() if len(v) > 0}
-            data = gramex.cache.query(text(query), engine, state, params=all_params)
+            data = gramex.cache.query(sa.text(query), engine, state, params=all_params)
             data = transform(data) if callable(transform) else data
             return _filter_frame(data, meta=meta, controls=controls, args=args)
         elif table:
@@ -246,16 +256,16 @@ def filter(url, args={}, meta={}, engine=None, table=None, ext=None,
         raise ValueError('engine: %s invalid. Can be sqlalchemy|file|dataframe' % engine)
 
 
-def delete(url, meta={}, args=None, engine=None, table=None, ext=None, id=None,
+def delete(url, meta={}, args=None, engine=None, table=None, ext=None, id=None, columns=None,
            query=None, queryfile=None, transform=None, transform_kwargs={}, **kwargs):
     '''
     Deletes data using URL query parameters. Typical usage::
 
         count = gramex.data.delete(dataframe, args=handler.args, id=['id'])
         count = gramex.data.delete('file.csv', args=handler.args, id=['id'])
-        count = gramex.data.delete('mysql://server/db', table='table', args=handler.args, id='id')
+        count = gramex.data.delete('mysql://server/db', table='x', args=handler.args, id=['id'])
 
-    ``id`` is a column name or a list of column names defining the primary key.
+    ``id`` is a list of column names defining the primary key.
     Calling this in a handler with ``?id=1&id=2`` deletes rows with id is 1 or 2.
 
     It accepts the same parameters as :py:func:`filter`, and returns the number
@@ -280,24 +290,23 @@ def delete(url, meta={}, args=None, engine=None, table=None, ext=None, id=None,
     elif engine == 'sqlalchemy':
         if table is None:
             raise ValueError('No table: specified')
-        engine = create_engine(url, **kwargs)
+        engine = alter(url, table, columns, **kwargs)
         return _filter_db(engine, table, meta=meta, controls=controls, args=args,
                           source='delete', id=id)
     else:
         raise ValueError('engine: %s invalid. Can be sqlalchemy|file|dataframe' % engine)
-    return 0
 
 
-def update(url, meta={}, args=None, engine=None, table=None, ext=None, id=None,
+def update(url, meta={}, args=None, engine=None, table=None, ext=None, id=None, columns=None,
            query=None, queryfile=None, transform=None, transform_kwargs={}, **kwargs):
     '''
     Update data using URL query parameters. Typical usage::
 
         count = gramex.data.update(dataframe, args=handler.args, id=['id'])
         count = gramex.data.update('file.csv', args=handler.args, id=['id'])
-        count = gramex.data.update('mysql://server/db', table='table', args=handler.args, id='id')
+        count = gramex.data.update('mysql://server/db', table='x', args=handler.args, id=['id'])
 
-    ``id`` is a column name or a list of column names defining the primary key.
+    ``id`` is a list of column names defining the primary key.
     Calling this in a handler with ``?id=1&x=2`` updates x=2 where id=1.
 
     It accepts the same parameters as :py:func:`filter`, and returns the number of updated rows.
@@ -321,24 +330,23 @@ def update(url, meta={}, args=None, engine=None, table=None, ext=None, id=None,
     elif engine == 'sqlalchemy':
         if table is None:
             raise ValueError('No table: specified')
-        engine = create_engine(url, **kwargs)
+        engine = alter(url, table, columns, **kwargs)
         return _filter_db(engine, table, meta=meta, controls=controls, args=args,
                           source='update', id=id)
     else:
         raise ValueError('engine: %s invalid. Can be sqlalchemy|file|dataframe' % engine)
-    return 0
 
 
-def insert(url, meta={}, args=None, engine=None, table=None, ext=None, id=None,
+def insert(url, meta={}, args=None, engine=None, table=None, ext=None, id=None, columns=None,
            query=None, queryfile=None, transform=None, transform_kwargs={}, **kwargs):
     '''
     Insert data using URL query parameters. Typical usage::
 
         count = gramex.data.insert(dataframe, args=handler.args, id=['id'])
         count = gramex.data.insert('file.csv', args=handler.args, id=['id'])
-        count = gramex.data.insert('mysql://server/db', table='table', args=handler.args, id='id')
+        count = gramex.data.insert('mysql://server/db', table='x', args=handler.args, id=['id'])
 
-    ``id`` is a column name or a list of column names defining the primary key.
+    ``id`` is a list of column names defining the primary key.
     Calling this in a handler with ``?id=3&x=2`` inserts a new record with id=3 and x=2.
 
     If the target file / table does not exist, it is created.
@@ -348,7 +356,9 @@ def insert(url, meta={}, args=None, engine=None, table=None, ext=None, id=None,
     if engine is None:
         engine = get_engine(url)
     _pop_controls(args)
-    meta.update({'filters': [], 'ignored': []})
+    if not args:
+        raise ValueError('No args: specified')
+    meta.update({'filters': [], 'ignored': [], 'inserted': []})
     # If values do not have equal number of elements, pad them and warn
     rowcount = max(len(val) for val in args.values())
     for key, val in args.items():
@@ -377,23 +387,48 @@ def insert(url, meta={}, args=None, engine=None, table=None, ext=None, id=None,
     elif engine == 'sqlalchemy':
         if table is None:
             raise ValueError('No table: specified')
-        engine = create_engine(url, **kwargs)
+        engine = alter(url, table, columns, **kwargs)
         try:
             cols = get_table(engine, table).columns
-        except sqlalchemy.exc.NoSuchTableError:
+        except sa.exc.NoSuchTableError:
             pass
         else:
             rows = _pop_columns(rows, [col.name for col in cols], meta['ignored'])
         if '.' in table:
             kwargs['schema'], table = table.rsplit('.', 1)
-        # pandas does not document engine.dialect.has_table so it might change.
+        # If the DB doesn't yet have the table, create it WITH THE PRIMARY KEYS.
+        # Note: pandas does not document engine.dialect.has_table so it might change.
         if not engine.dialect.has_table(engine, table) and id:
             engine.execute(pd.io.sql.get_schema(rows, name=table, keys=id, con=engine))
-        rows.to_sql(table, engine, if_exists='append', index=False, **kwargs)
+
+        def insert_method(tbl, conn, keys, data_iter):
+            '''Pandas .to_sql() doesn't return inserted row primary keys. Capture it in meta'''
+            data = [dict(zip(keys, row)) for row in data_iter]
+            # If the ?id= is not provided, Pandas creates a schema based on available columns,
+            # without the `id` column. SQLAlchemy won't return inserted_primary_key unless the
+            # metadata has a primary key. So, hoping that the table already has a primary key,
+            # load table from DB via extend_existing=True.
+            sa_table = sa.Table(table, tbl.table.metadata,
+                                extend_existing=True, autoload_with=engine)
+            r = conn.execute(sa_table.insert(), data)
+            # SQLAlchemy 1.4+ supports inserted_primary_key_rows, but is in beta (Nov 2020).
+            # ids = getattr(r, 'inserted_primary_key_rows', [])
+            # If we have SQLAlchemy 1.3, only single inserts have an inserted_primary_key.
+            ids = [r.inserted_primary_key] if hasattr(r, 'inserted_primary_key') else []
+            # Add non-empty IDs as a dict with associated keys
+            id_cols = [col.name for col in sa_table.primary_key]
+            for row in ids:
+                if row:
+                    meta['inserted'].append(dict(zip(id_cols, row)))
+
+        kwargs['method'] = insert_method
+        # If user passes ?col= with empty value, replace with NULL. If the column is an INT/FLOAT,
+        # type conversion int('') / float('') will fail.
+        rows.replace('', None, inplace=True)
+        pd.io.sql.to_sql(rows, table, engine, if_exists='append', index=False, **kwargs)
         return len(rows)
     else:
         raise ValueError('engine: %s invalid. Can be sqlalchemy|file|dataframe' % engine)
-    return 0
 
 
 def get_engine(url):
@@ -411,13 +446,13 @@ def get_engine(url):
     if isinstance(url, pd.DataFrame):
         return 'dataframe'
     try:
-        url = sqlalchemy.engine.url.make_url(url)
-    except sqlalchemy.exc.ArgumentError:
+        url = sa.engine.url.make_url(url)
+    except sa.exc.ArgumentError:
         return 'dir' if os.path.isdir(url) else 'file'
     try:
         url.get_driver_name()
         return 'sqlalchemy'
-    except sqlalchemy.exc.NoSuchModuleError:
+    except sa.exc.NoSuchModuleError:
         return url.drivername
 
 
@@ -431,20 +466,18 @@ def create_engine(url, **kwargs):
     re-use the engine objects within this module.
     '''
     if url not in _ENGINE_CACHE:
-        _ENGINE_CACHE[url] = sqlalchemy.create_engine(url, **kwargs)
+        _ENGINE_CACHE[url] = sa.create_engine(url, **kwargs)
     return _ENGINE_CACHE[url]
 
 
-def get_table(engine, table):
+def get_table(engine, table, **kwargs):
     '''Return the sqlalchemy table from the engine and table name'''
     if engine not in _METADATA_CACHE:
-        _METADATA_CACHE[engine] = sqlalchemy.MetaData()
+        _METADATA_CACHE[engine] = sa.MetaData(bind=engine)
     metadata = _METADATA_CACHE[engine]
     if '.' in table:
-        schema, tbl = table.rsplit('.', 1)
-        return sqlalchemy.Table(tbl, metadata, autoload=True, autoload_with=engine, schema=schema)
-    else:
-        return sqlalchemy.Table(table, metadata, autoload=True, autoload_with=engine)
+        kwargs['schema'], table = table.rsplit('.', 1)
+    return sa.Table(table, metadata, autoload=True, autoload_with=engine, **kwargs)
 
 
 def _pop_controls(args):
@@ -785,11 +818,11 @@ def _filter_db(engine, table, meta, controls, args, source='select', id=[]):
     colslist = cols.keys()
 
     if source == 'delete':
-        query = sqlalchemy.delete(table)
+        query = sa.delete(table)
     elif source == 'update':
-        query = sqlalchemy.update(table)
+        query = sa.update(table)
     else:
-        query = sqlalchemy.select([table])
+        query = sa.select([table])
     cols_for_update = {}
     cols_having = []
     for key, vals in args.items():
@@ -844,7 +877,7 @@ def _filter_db(engine, table, meta, controls, args, source='select', id=[]):
                     # Convert aggregation into SQLAlchemy query
                     agg = agg.lower()
                     typ[key] = _agg_type.get(agg, cols[col].type.python_type)
-                    agg_func = getattr(sqlalchemy.sql.expression.func, agg)
+                    agg_func = getattr(sa.sql.expression.func, agg)
                     agg_cols[key] = agg_func(cols[col]).label(key)
             if not agg_cols:
                 return pd.DataFrame()
@@ -864,7 +897,7 @@ def _filter_db(engine, table, meta, controls, args, source='select', id=[]):
             meta['sort'], ignore_sorts = _filter_sort_columns(
                 controls['_sort'], colslist + query.columns.keys())
             for col, asc in meta['sort']:
-                orderby = sqlalchemy.asc if asc else sqlalchemy.desc
+                orderby = sa.asc if asc else sa.desc
                 query = query.order_by(orderby(col))
             if len(ignore_sorts) > 0:
                 meta['ignored'].append(('_sort', ignore_sorts))
@@ -1090,9 +1123,9 @@ def dirstat(url, timeout=10, **kwargs):
         - ``level``: path depth (i.e. the number of paths in dir)
     '''
     try:
-        url = sqlalchemy.engine.url.make_url(url)
+        url = sa.engine.url.make_url(url)
         target = url.database
-    except sqlalchemy.exc.ArgumentError:
+    except sa.exc.ArgumentError:
         target = url
     if not os.path.isdir(target):
         raise OSError('dirstat: %s is not a directory' % target)
@@ -1231,3 +1264,81 @@ def filtercols(url, args={}, meta={}, engine=None, table=None, ext=None,
         col_args['_limit'] = [limit]
         result[col] = gramex.data.filter(url, table=table, args=col_args, **kwargs)
     return result
+
+
+def alter(url: str, table: str, columns: dict = None, **kwargs):
+    '''
+    Create or alter a table with columns specified::
+
+        gramex.data.alter(url, table, columns={
+            'id': {'type': 'int', 'primary_key': True, 'autoincrement': True},
+            'email': {'nullable': True, 'default': 'none'},
+            'age': {'type': 'float', 'nullable': False, 'default': 18},
+        })
+
+    It accepts the following parameters:
+
+    :arg str url: sqlalchemy URL
+    :arg str table: table name
+    :arg dict columns: column names, with values are SQL types, or dicts with keys:
+        - ``type`` (str), e.g. ``"VARCHAR(10)"``
+        - ``default`` (str/int/float/bool), e.g. ``"none@example.org"``
+        - ``nullable`` (bool), e.g. ``False``
+        - ``primary_key`` (bool), e.g. ``True`` -- used only when creating new tables
+        - ``autoincrement`` (bool), e.g. ``True`` -- used only when creating new tables
+    :return: sqlalchemy engine
+
+    Other kwargs are passed to ``sqlalchemy.create_engine()``.
+
+    If the table exists, any new columns are added. Existing columns are unchanged.
+
+    If the table does not exist, the table is created with the specified columns.
+
+    Note: ``primary_key`` and ``autoincrement`` don't work on existing tables because:
+        - SQLite disallows PRIMARY KEY with ALTER. https://stackoverflow.com/a/1120030/100904
+        - AUTO_INCREMENT doesn't work without PRIMARY KEY in MySQL
+    '''
+    engine = create_engine(url, **kwargs)
+    if columns is None:
+        return engine
+    try:
+        db_table = get_table(engine, table)
+    except sa.exc.NoSuchTableError:
+        # If the table's not in the DB, create it
+        cols = []
+        for name, row in columns.items():
+            row = dict({'type': row} if isinstance(row, str) else row, name=name)
+            col_type = row.get('type', 'text')
+            if isinstance(col_type, str):
+                # Use eval() to handle direct types like INTEGER *and* expressions like VARCHAR(3)
+                row['type'] = eval(col_type.upper(), vars(sa.types))    # nosec
+            row['type_'] = row.pop('type')
+            if 'default' in row:
+                row['server_default'] = str(row.pop('default'))
+            cols.append(sa.Column(**row))
+        sa.Table(table, _METADATA_CACHE[engine], *cols).create(engine)
+    else:
+        quote = engine.dialect.identifier_preparer.quote_identifier
+        # If the table's already in the DB, add new columns. We can't change column types
+        with engine.connect() as conn:
+            with conn.begin():
+                for name, row in columns.items():
+                    if name in db_table.columns:
+                        continue
+                    row = {'type': row} if isinstance(row, str) else row
+                    col_type = row.get('type', 'text')
+                    constraints = []
+                    if 'nullable' in row:
+                        constraints.append('' if row['nullable'] else 'NOT NULL')
+                    if 'default' in row:
+                        # repr() converts int, float properly,
+                        #   str into 'str' with single quotes (which is the MySQL standard)
+                        #   TODO: datetime and other types will fail
+                        constraints += ['DEFAULT', repr(row['default'])]
+                    # This syntax works on DB2, MySQL, Oracle, PostgreSQL, SQLite
+                    conn.execute(
+                        f'ALTER TABLE {quote(table)} '
+                        f'ADD COLUMN {quote(name)} {col_type} {" ".join(constraints)}')
+        # Refresh table metadata after altering
+        get_table(engine, table, extend_existing=True)
+    return engine
