@@ -3,20 +3,20 @@ import os
 import six
 import time
 import logging
-import datetime
 import mimetypes
 import traceback
 import tornado.gen
 import gramex.cache
+from typing import Union
 from binascii import b2a_base64, hexlify
 from orderedattrdict import AttrDict
-from six.moves.urllib_parse import urlparse, urlsplit, urljoin, urlencode
+from urllib.parse import urlsplit, urlunsplit, parse_qsl, urljoin, urlencode
 from tornado.web import RequestHandler, HTTPError, MissingArgumentError, decode_signed_value
 from tornado.websocket import WebSocketHandler
 from gramex import conf, __version__
 from gramex.config import merge, objectpath, app_log
-from gramex.transforms import build_transform, CacheLoader
-from gramex.http import UNAUTHORIZED, FORBIDDEN, BAD_REQUEST
+from gramex.transforms import build_transform, build_log_info, CacheLoader
+from gramex.http import UNAUTHORIZED, FORBIDDEN, BAD_REQUEST, METHOD_NOT_ALLOWED
 from gramex.cache import get_store
 # We don't use these, but these stores used to be defined here. Programs may import these
 from gramex.cache import KeyStore, JSONStore, HDF5Store, SQLiteStore, RedisStore    # noqa
@@ -29,7 +29,7 @@ _arg_default = object()
 
 class BaseMixin(object):
     @classmethod
-    def setup(cls, transform={}, redirect={}, auth=None, log=None, set_xsrf=None,
+    def setup(cls, transform={}, redirect={}, methods=None, auth=None, log=None, set_xsrf=None,
               error=None, xsrf_cookies=None, **kwargs):
         '''
         One-time setup for all request handlers. This is called only when
@@ -50,13 +50,14 @@ class BaseMixin(object):
         cls.setup_error(error)
         cls.setup_xsrf(xsrf_cookies)
         cls.setup_log()
+        cls.setup_httpmethods(methods)
 
         # app.settings.debug enables debugging exceptions using pdb
         if conf.app.settings.get('debug', False):
             cls.log_exception = cls.debug_exception
 
     # A list of special keys for BaseHandler. Can be extended by other classes.
-    special_keys = ['transform', 'redirect', 'auth', 'log', 'set_xsrf',
+    special_keys = ['transform', 'redirect', 'methods', 'auth', 'log', 'set_xsrf',
                     'error', 'xsrf_cookies', 'headers']
 
     @classmethod
@@ -71,6 +72,18 @@ class BaseMixin(object):
             kwargs.pop(special_key, None)
         return kwargs
 
+    @classmethod
+    def setup_httpmethods(cls, methods: Union[list, tuple, str]):
+        if methods is None:
+            return
+        if isinstance(methods, (list, tuple)):
+            methods = ' '.join(methods)
+        if not isinstance(methods, str):
+            raise ValueError('methods: %r invalid -- use a string/list, e.g. [GET, PUT]')
+        if methods:
+            cls._http_methods = set(methods.upper().replace(',', ' ').split())
+            cls._on_init_methods.append(cls.check_http_method)
+
     @classmethod
     def setup_default_kwargs(cls):
         '''
@@ -144,6 +157,7 @@ class BaseMixin(object):
         cls._session_store = session_store_cache[key]
         cls.session = property(cls.get_session)
         cls._session_expiry = session_conf.get('expiry')
+        cls._session_cookie_id = session_conf.get('cookie', 'sid')
         cls._session_cookie = {
             key: session_conf[key] for key in ('domain', 'httponly', 'secure')
             if key in session_conf
@@ -184,7 +198,7 @@ class BaseMixin(object):
         this same consistent way.
         '''
         # Ensure that redirect is a dictionary before proceeding.
-        if isinstance(redirect, six.string_types):
+        if isinstance(redirect, str):
             redirect = {'url': redirect}
         if not isinstance(redirect, dict):
             app_log.error('url:%s.redirect must be a URL or a dict, not %s',
@@ -207,7 +221,7 @@ class BaseMixin(object):
                 def redirect_method(handler):
                     next_uri = method(handler)
                     if next_uri is not None:
-                        target = urlparse(next_uri)
+                        target = urlsplit(next_uri)
                         if not target.scheme and not target.netloc:
                             return next_uri
                         req = handler.request
@@ -227,7 +241,7 @@ class BaseMixin(object):
             auth = AttrDict()
         # Set up the auth
         if isinstance(auth, dict):
-            cls._login_url = auth.get('login_url', None)
+            cls._auth = auth
             cls._on_init_methods.append(cls.authorize)
             cls.permissions = []
             # Add check for condition
@@ -341,6 +355,12 @@ class BaseMixin(object):
         '''
         cls.check_xsrf_cookie = cls.noop if xsrf_cookies is False else cls.xsrf_ajax
 
+    def check_http_method(self):
+        '''If method: [...] is specified, reject all methods not in the allowed methods set'''
+        if self.request.method not in self._http_methods:
+            raise HTTPError(METHOD_NOT_ALLOWED, f'{self.name}: method {self.request.method} ' +
+                            f'not in allowed methods {self._http_methods}')
+
     def xsrf_ajax(self):
         '''
         TODO: explain things clearly.
@@ -364,7 +384,7 @@ class BaseMixin(object):
         for method in self.redirects:
             next_url = method(self)
             if next_url:
-                self.session['_next_url'] = urljoin(self.request.uri, next_url)
+                self.session['_next_url'] = urljoin(self.xrequest_uri, next_url)
                 return
         self.session.setdefault('_next_url', '/')
 
@@ -418,7 +438,7 @@ class BaseMixin(object):
                 self._write_headers(headers.items())
                 # result may be a generator / list from build_transform,
                 # or a str/bytes/unicode from Template.generate. Handle both
-                if isinstance(result, (six.string_types, six.binary_type)):
+                if isinstance(result, (str, bytes)):
                     self.write(result)
                 else:
                     for item in result:
@@ -449,7 +469,7 @@ class BaseMixin(object):
             kwargs['secure'] = True
         # Websockets cannot set cookies. They raise a RuntimeError. Ignore those.
         try:
-            self.set_secure_cookie('sid', session_id, **kwargs)
+            self.set_secure_cookie(self._session_cookie_id, session_id, **kwargs)
         except RuntimeError:
             pass
         return session_id
@@ -486,7 +506,7 @@ class BaseMixin(object):
         if getattr(self, '_session', None) is None:
             # Populate self._session based on the sid. If there's no sid cookie,
             # generate one and create an associated session object
-            session_id = self.get_secure_cookie('sid', max_age_days=9999999)
+            session_id = self.get_secure_cookie(self._session_cookie_id, max_age_days=9999999)
             # If there's no session id cookie "sid", create a random 32-char cookie
             if session_id is None:
                 session_id = self._set_new_session_id(expires_days)
@@ -573,7 +593,7 @@ class BaseMixin(object):
         '''
         # For efficiency reasons, don't call get_session every time. Check
         # session only if there's a valid sid cookie (with possibly long expiry)
-        if self.get_secure_cookie('sid', max_age_days=9999999):
+        if self.get_secure_cookie(self._session_cookie_id, max_age_days=9999999):
             session = self.get_session()
             if '_i' in session:
                 session['_l'] = time.time()
@@ -590,7 +610,7 @@ class BaseHandler(RequestHandler, BaseMixin):
         # Convert this to proper unicode using UTF-8 and store in self.args
         self.args = {}
         for k in self.request.arguments:
-            key = (k if isinstance(k, six.binary_type) else k.encode('latin-1')).decode('utf-8')
+            key = (k if isinstance(k, bytes) else k.encode('latin-1')).decode('utf-8')
             # Invalid unicode (e.g. ?x=%f4) throws HTTPError. This disrupts even
             # error handlers. So if there's invalid unicode, log & continue.
             try:
@@ -633,6 +653,15 @@ class BaseHandler(RequestHandler, BaseMixin):
         return self.args[name][0 if first else -1]
 
     def prepare(self):
+        # If X-Request-URI is specified, use it. Else, when redirecting use RELATIVE URL. That
+        # allows nginx to proxy_redirect Location headers
+        self.xrequest_uri = self.request.headers.get('X-Request-URI', self.request.uri)
+        # When passing to URL query parameters (e.g. /login/?next=), use the full URL
+        self.xrequest_full_url = urljoin(self.request.full_url(), self.xrequest_uri)
+        # For third-party redirection (e.g. Google Auth / Twitter needs a callback URI), then use
+        # the full URL WITHOUT query parameters
+        self.xredirect_uri = '{0.scheme:s}://{0.netloc:s}{0.path:s}'.format(
+            urlsplit(self.xrequest_full_url))
         for method in self._on_init_methods:
             method(self)
 
@@ -665,21 +694,17 @@ class BaseHandler(RequestHandler, BaseMixin):
         if not self.current_user:
             # Redirect non-AJAX requests GET/HEAD to login URL (if it's a string)
             ajax = self.request.headers.get('X-Requested-With', '').lower() == 'xmlhttprequest'
-            if self.request.method in ('GET', 'HEAD') and not ajax:
-                url = self.get_login_url() if self._login_url is None else self._login_url
-                # If login_url is a string, redirect
-                if isinstance(url, six.string_types):
-                    if '?' not in url:
-                        if urlsplit(url).scheme:
-                            # if login url is absolute, make next absolute too
-                            next_url = self.request.full_url()
-                        else:
-                            next_url = self.request.uri
-                        next_key = 'next'
-                        if isinstance(self.conf.kwargs.auth, dict):
-                            next_key = self.conf.kwargs.auth.get('query', 'next')
-                        url += '?' + urlencode({next_key: next_url})
-                    self.redirect(url)
+            if not ajax and self.request.method in ('GET', 'HEAD'):
+                auth = getattr(self, '_auth', {})
+                url = auth.get('login_url', self.get_login_url())
+                # If login_url: false, don't redirect to a login URL. Only redirect if it's a URL
+                if isinstance(url, str):
+                    # Redirect to the login_url adding ?next=<X-Request-URI>
+                    p = urlsplit(url)
+                    q = parse_qsl(p.query)
+                    q.append((auth.get('query', 'next'), self.xrequest_full_url))
+                    target = urlunsplit((p.scheme, p.netloc, p.path, urlencode(q), p.fragment))
+                    self.redirect(target)
                     return
             # Else, send a 401 header
             raise HTTPError(UNAUTHORIZED)
@@ -768,8 +793,8 @@ class BaseHandler(RequestHandler, BaseMixin):
         '''
         result = AttrDict()
 
-        args_type = six.text_type
-        if len(args) > 0 and args[0] in (six.text_type, six.binary_type, list, None):
+        args_type = str
+        if len(args) > 0 and args[0] in (str, bytes, list, None):
             args_type, args = args[0], args[1:]
 
         for key in args:
@@ -831,7 +856,7 @@ class BaseHandler(RequestHandler, BaseMixin):
             for key, val in self.args.items():
                 if key not in args and key not in kwargs:
                     result[key] = val
-        elif args_type in (six.string_types, six.binary_type):
+        elif args_type in (str, bytes):
             for key, val in self.args.items():
                 if key not in args and key not in kwargs:
                     result[key] = args_type(val[0])
@@ -941,63 +966,5 @@ handle_cache = {}
 def _handle(path):
     '''Returns a cached append-binary handle to path'''
     if path not in handle_cache:
-        # In Python 2, csv writerow writes byte string. In PY3, it's to a unicode string.
-        # Open file handles accordingly
-        handle_cache[path] = open(path, 'ab') if six.PY2 else io.open(path, 'a', encoding='utf-8')
+        handle_cache[path] = io.open(path, 'a', encoding='utf-8')
     return handle_cache[path]
-
-
-def build_log_info(keys, *vars):
-    '''
-    Creates a ``handler.method(vars)`` that returns a dictionary of computed
-    values. ``keys`` defines what keys are returned in the dictionary. The values
-    are computed using the formulas in the code.
-    '''
-    # Define direct keys. These can be used as-is
-    direct_vars = {
-        'name': 'handler.name',
-        'class': 'handler.__class__.__name__',
-        'time': 'round(time.time() * 1000, 0)',
-        'datetime': 'datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%SZ")',
-        'method': 'handler.request.method',
-        'uri': 'handler.request.uri',
-        'ip': 'handler.request.remote_ip',
-        'status': 'handler.get_status()',
-        'duration': 'round(handler.request.request_time() * 1000, 0)',
-        'port': 'conf.app.listen.port',
-        # TODO: get_content_size() is not available in RequestHandler
-        # 'size': 'handler.get_content_size()',
-        'user': '(handler.current_user or {}).get("id", "")',
-        'session': 'handler.session.get("id", "")',
-        'error': 'getattr(handler, "_exception", "")',
-    }
-    # Define object keys for us as key.value. E.g. cookies.sid, user.email, etc
-    object_vars = {
-        'args': 'handler.get_argument("{val}", "")',
-        'request': 'getattr(handler.request, "{val}", "")',
-        'headers': 'handler.request.headers.get("{val}", "")',
-        'cookies': 'handler.request.cookies["{val}"].value ' +
-                   'if "{val}" in handler.request.cookies else ""',
-        'user': '(handler.current_user or {{}}).get("{val}", "")',
-        'env': 'os.environ.get("{val}", "")',
-    }
-    vals = []
-    for key in keys:
-        if key in vars:
-            vals.append('"{}": {},'.format(key, key))
-            continue
-        if key in direct_vars:
-            vals.append('"{}": {},'.format(key, direct_vars[key]))
-            continue
-        if '.' in key:
-            prefix, value = key.split('.', 2)
-            if prefix in object_vars:
-                vals.append('"{}": {},'.format(key, object_vars[prefix].format(val=value)))
-                continue
-        app_log.error('Skipping unknown key %s', key)
-    code = compile('def fn(handler, %s):\n\treturn {%s}' % (', '.join(vars), ' '.join(vals)),
-                   filename='log', mode='exec')
-    context = {'os': os, 'time': time, 'datetime': datetime, 'conf': conf, 'AttrDict': AttrDict}
-    # The code is constructed entirely by this function. Using exec is safe
-    exec(code, context)         # nosec
-    return context['fn']
