@@ -1,7 +1,6 @@
-import six
 import tornado.web
 import tornado.gen
-from six.moves.urllib_parse import urlsplit, urlunsplit, parse_qs, urlencode
+from urllib.parse import urlsplit, urlunsplit, parse_qs, urlencode
 from tornado.httputil import HTTPHeaders
 from tornado.httpclient import AsyncHTTPClient, HTTPRequest
 from gramex.transforms import build_transform
@@ -30,7 +29,6 @@ class ProxyHandler(BaseHandler, BaseWebSocketHandler):
         - Any string value is formatted with ``handler`` as a variable.
     :arg dict default: Default URL query parameters
     :arg dict headers: HTTP headers to set on the response
-    :arg list methods: list of HTTP methods allowed (default: [GET, HEAD, POST])
     :arg function prepare: A function that accepts any of ``handler`` and ``request``
         (a tornado.httpclient.HTTPRequest) and modifies the ``request`` in-place
     :arg function modify: A function that accepts any of ``handler``, ``request``
@@ -62,8 +60,7 @@ class ProxyHandler(BaseHandler, BaseWebSocketHandler):
     '''
     @classmethod
     def setup(cls, url, request_headers={}, default={}, prepare=None, modify=None,
-              headers={}, methods=['GET', 'HEAD', 'POST'],
-              connect_timeout=20, request_timeout=20, **kwargs):
+              headers={}, connect_timeout=20, request_timeout=20, **kwargs):
         super(ProxyHandler, cls).setup(**kwargs)
         WebSocketHandler._setup(cls, **kwargs)
         cls.url, cls.request_headers, cls.default = url, request_headers, default
@@ -76,8 +73,7 @@ class ProxyHandler(BaseHandler, BaseWebSocketHandler):
                     {'function': fn}, filename='url:%s.%s' % (cls.name, key),
                     vars={'handler': None, 'request': None, 'response': None})
         cls.browser = AsyncHTTPClient()
-        for method in methods:
-            setattr(cls, method.lower(), cls.method)
+        cls.post = cls.put = cls.delete = cls.patch = cls.options = cls.get
 
     def authorize(self, *args, **kwargs):
         if self.request.headers.get('Upgrade', '') == 'websocket':
@@ -86,7 +82,7 @@ class ProxyHandler(BaseHandler, BaseWebSocketHandler):
             super(ProxyHandler, self).authorize()
 
     @tornado.gen.coroutine
-    def method(self, *path_args):
+    def get(self, *path_args):
         ws = self.request.headers.get('Upgrade', '') == 'websocket'
         if ws:
             return WebSocketHandler.get(self)
@@ -99,14 +95,14 @@ class ProxyHandler(BaseHandler, BaseWebSocketHandler):
                 if key in self.request.headers:
                     headers[key] = self.request.headers[key]
             else:
-                headers[key] = six.text_type(val).format(handler=self)
+                headers[key] = str(val).format(handler=self)
 
         # Update query parameters
         # TODO: use a named capture for path_args? This is not the right method
         parts = urlsplit(self.url.format(*path_args))
         params = {
-            key: ([six.text_type(v).format(handler=self) for v in val] if isinstance(val, list)
-                  else six.text_type(val).format(handler=self))
+            key: ([str(v).format(handler=self) for v in val] if isinstance(val, list)
+                  else str(val).format(handler=self))
             for key, val in self.default.items()
         }
         params.update(parse_qs(parts.query))
