@@ -1,5 +1,4 @@
 import os
-import six
 import json
 import gramex
 import tornado.gen
@@ -17,19 +16,13 @@ custom_responses = {
 store_cache = {}
 
 
-class SocialMixin(object):
+class SocialHandler(BaseHandler):
     @classmethod
-    def setup_social(cls, user_info, transform={}, methods=['get', 'post'], **kwargs):
+    def setup_social(cls, user_info, **kwargs):
         # Session key that stores the user info
         cls.user_info = user_info
-
         # Set up methods
-        if not isinstance(methods, list):
-            methods = [methods]
-        methods = set(method.lower().strip() for method in methods)
-        for method in ('get', 'post', 'put', 'patch'):
-            if method in methods:
-                setattr(cls, method, cls.run)
+        cls.post = cls.put = cls.delete = cls.patch = cls.options = cls.get
 
     @tornado.gen.coroutine
     def social_response(self, response):
@@ -53,7 +46,7 @@ class SocialMixin(object):
         if content and response.code == OK:
             content = yield gramex.service.threadpool.submit(self.run_transforms, content=content)
         # Convert to JSON if required
-        if not isinstance(content, (six.binary_type, six.text_type)):
+        if not isinstance(content, (str, bytes)):
             content = json.dumps(content, ensure_ascii=True, separators=(',', ':'))
         raise tornado.gen.Return(content)
 
@@ -125,7 +118,7 @@ class SocialMixin(object):
         return token
 
 
-class TwitterRESTHandler(SocialMixin, BaseHandler, TwitterMixin):
+class TwitterRESTHandler(SocialHandler, TwitterMixin):
     '''
     Proxy for the Twitter 1.1 REST API via these ``kwargs``::
 
@@ -160,7 +153,7 @@ class TwitterRESTHandler(SocialMixin, BaseHandler, TwitterMixin):
         cls.setup_social('user.twitter', **kwargs)
 
     @tornado.gen.coroutine
-    def run(self, path=None):
+    def get(self, path=None):
         path = self.kwargs.get('path', path)
         if not path and self.request.method == 'GET':
             yield self.login()
@@ -195,15 +188,14 @@ class TwitterRESTHandler(SocialMixin, BaseHandler, TwitterMixin):
             self.redirect_next()
         else:
             self.save_redirect_page()
-            yield self.authorize_redirect(callback_uri=self.request.protocol + "://" +
-                                          self.request.host + self.request.uri)
+            yield self.authorize_redirect(callback_uri=self.xredirect_uri)
 
     def _oauth_consumer_token(self):
         return dict(key=self.kwargs['key'],
                     secret=self.kwargs['secret'])
 
 
-class FacebookGraphHandler(SocialMixin, BaseHandler, FacebookGraphMixin):
+class FacebookGraphHandler(SocialHandler, FacebookGraphMixin):
     '''
     Proxy for the Facebook Graph API via these ``kwargs``::
 
@@ -238,7 +230,7 @@ class FacebookGraphHandler(SocialMixin, BaseHandler, FacebookGraphMixin):
         cls.setup_social('user.facebook', **kwargs)
 
     @tornado.gen.coroutine
-    def run(self, path=None):
+    def get(self, path=None):
         path = self.kwargs.get('path', path)
         if not path and self.request.method == 'GET':
             yield self.login()
@@ -255,10 +247,9 @@ class FacebookGraphHandler(SocialMixin, BaseHandler, FacebookGraphMixin):
 
     @tornado.gen.coroutine
     def login(self):
-        redirect_uri = self.request.protocol + "://" + self.request.host + self.request.uri
         if self.get_argument('code', False):
             info = self.session[self.user_info] = yield self.get_authenticated_user(
-                redirect_uri=redirect_uri,
+                redirect_uri=self.xredirect_uri,
                 client_id=self.kwargs['key'],
                 client_secret=self.kwargs['secret'],
                 code=self.get_argument('code'))
@@ -269,6 +260,6 @@ class FacebookGraphHandler(SocialMixin, BaseHandler, FacebookGraphMixin):
             self.save_redirect_page()
             scope = self.kwargs.get('scope', 'user_posts,read_insights')
             yield self.authorize_redirect(
-                redirect_uri=redirect_uri,
+                redirect_uri=self.xredirect_uri,
                 client_id=self.kwargs['key'],
                 extra_params={'scope': scope})
