@@ -136,11 +136,8 @@ def apply_commands(rule: Dict[str, dict], shapes, data: dict):
     for pattern, spec in rule.items():
         if pattern in rule_cmdlist:
             continue
-        shape_matched = False
-        for shape in shapes:
-            if not fnmatchcase(shape.name, pattern):
-                continue
-            shape_matched = True
+        matched_shapes = [shape for shape in shapes if fnmatchcase(shape.name, pattern)]
+        for shape in matched_shapes:
             # Clone all slides into the `clones` list BEFORE applying any command. Ensures that
             # commands applied to the shape don't propagate into its clones
             clones = []
@@ -172,7 +169,7 @@ def apply_commands(rule: Dict[str, dict], shapes, data: dict):
                 if is_group:
                     apply_commands(spec, SlideShapes(clone.shape.element, shapes), shape_data)
         # Warn if the pattern is neither a shape nor a command
-        if (not shape_matched and pattern not in special_cmdlist and
+        if (not matched_shapes and pattern not in special_cmdlist and
                 pattern not in commands.cmdlist):
             app_log.warn('pptgen2: No shape matches pattern: %s', pattern)
 
@@ -478,6 +475,11 @@ def commandline(args=None):
     gramex.config.merge(rules, args)
     rules.setdefault('target', 'output.pptx')
     rules.setdefault('mode', 'expr')
+    # Allow importing python files in current directory
+    sys.path.append('.')
+    # Generate output
     gramex.pptgen2.pptgen(**rules)
-    if not rules.get('no-open', False):
+    # If --no-open is specified, or the OS doesn't have startfile (e.g. Linux), stop here.
+    # Otherwise, open the output PPTX created
+    if not rules.get('no-open', False) and hasattr(os, 'startfile'):
         os.startfile(rules['target'])
