@@ -11,20 +11,23 @@ import io
 import matplotlib.cm
 import matplotlib.colors
 import os
+import pandas as pd
 import pptx
 import pptx.util
 import re
 import requests
 from PIL import Image
 from functools import partial
-from gramex.config import objectpath
+from gramex.config import app_log, objectpath
 from gramex.transforms import build_transform
 from lxml.html import fragments_fromstring, builder, HtmlElement
+from orderedattrdict import AttrDict
+from pptx.chart import data as pptxchartdata
 from pptx.dml.color import RGBColor
 from pptx.dml.fill import FillFormat
 from pptx.enum.base import EnumValue
 from pptx.enum.dml import MSO_THEME_COLOR, MSO_FILL
-from pptx.enum.text import PP_ALIGN
+from pptx.enum.text import PP_ALIGN, MSO_VERTICAL_ANCHOR
 from pptx.opc.constants import RELATIONSHIP_TYPE as RT
 from pptx.oxml.ns import _nsmap, qn
 from pptx.oxml.simpletypes import ST_Percentage
@@ -56,20 +59,23 @@ def expr(val, data: dict = {}):
     return val
 
 
-def assign(convert, *path: List[str]):
-    '''assign(int, 'font', 'size') returns a method(para, '10') -> para.font.size = int(10)'''
+def assign(convert, path: str):
+    '''assign(int, 'font.3.size') returns a method(para, '10') -> para.font[3].size = int(10)'''
+    path = path.split('.')
+
     def method(node, value, data):
         for p in path[:-1]:
-            node = getattr(node, p)
+            node = node[int(p)] if p.isdigit() else getattr(node, p)
         # To clear the bold, italic, etc, we need to set it to None. So don't convert None
         setattr(node, path[-1], None if value is None else convert(value))
+
     return method
 
 
 # Length utilities
 # ---------------------------------------------------------------------
 length_unit = pptx.util.Inches          # The default length unit is inches. This is exposed
-_length_expr = re.compile(r'''          # A length expression can be:
+_length_expr = re.compile(r'''          # A length expression can be:ex
     ([\d\.]+)                           #   Any integer or floating point (without + or -)
     \s*                                 #   optionally followed by spaces
     ("|in|inch|inches|cm|mm|pt|cp|centipoint|centipoints|emu|)  # and a unit that may be blank
@@ -82,7 +88,7 @@ length_alias = {'"': 'inches', 'in': 'inches', 'inch': 'inches',
 
 def length_class(unit_str: str):
     '''Converts unit string like in, inch, cm, etc to pptx.util.Inches, pptx.util.Cm, etc.'''
-    unit = length_alias.get(unit_str, unit_str).title()
+    unit = length_alias.get(unit_str.lower(), unit_str.lower()).title()
     if hasattr(pptx.util, unit):
         return getattr(pptx.util, unit)
     elif not unit:
@@ -97,7 +103,7 @@ def length(val: Union[str, int, float]) -> pptx.util.Length:
         match = _length_expr.match(val)
         if match:
             return length_class(match.group(2))(float(match.group(1)))
-    elif isinstance(val, (int, float)):
+    elif isinstance(val, (int, float, pd.np.number)):
         return length_unit(val)
     raise ValueError('Invalid length: %r' % val)
 
@@ -119,7 +125,7 @@ def fill_color(fill: FillFormat, val: Union[str, tuple, list, None]) -> None:
 
     - a named color, like ``black``
     - a hex value, like ``#f80`` or ``#ff8800``
-    - an RGB value, like ``rgb(255, 255, 0)``
+    - an RGB value, like ``rgb(255, 255, 0)`` or ``rgb(1, 0.5, 0.1)``
     - a tuple or list of RGB values, like ``(255, 255, 0)`` or ``[255, 255, 0]``
     - a theme color, like ``ACCENT_1``, ``ACCENT_2``, ``BACKGROUND_1``, ``DARK_1``, ``LIGHT_2``
     - a theme color with a brightness modifier, like ``ACCENT_1+40``, which is 40% brighter than
@@ -130,7 +136,11 @@ def fill_color(fill: FillFormat, val: Union[str, tuple, list, None]) -> None:
     if val == 'none':
         fill.background()
     elif isinstance(val, (list, tuple)):
-        fill.fore_color.rgb = RGBColor(*val[:3])
+        val = val[:3]
+        if any(isinstance(v, float) for v in val) and all(0 <= v <= 1 for v in val):
+            fill.fore_color.rgb = RGBColor(*(int(v * 256 if v < 1 else 255) for v in val))  # noqa
+        else:
+            fill.fore_color.rgb = RGBColor(*val)
     elif isinstance(val, str):
         val = color_map.get(val, val).upper()
         if val.startswith('#'):
@@ -179,9 +189,9 @@ def binary(val: Union[str, int, float]) -> bool:
     raise ValueError('Invalid boolean: %r' % val)
 
 
-def alignment(val: str) -> EnumValue:
-    '''Convert string to alignment value'''
-    alignment = getattr(PP_ALIGN, val.upper(), None)
+def alignment(enum, val: str) -> EnumValue:
+    '''Convert string to alignment value. enum can be PP_ALIGN or MSO_VERTICAL_ANCHOR'''
+    alignment = getattr(enum, val.upper(), None)
     if alignment is not None:
         return alignment
     raise ValueError('Invalid alignment: %s' % val)
@@ -250,14 +260,14 @@ def zoom(shape, spec, data: dict):
         shape.height = int(shape.height * val)
 
 
-def set_color(prop, attr, shape, spec, data: dict):
+def set_color(attr, shape, spec, data: dict):
     '''Generator for fill, stroke commands'''
     val = expr(spec, data)
     if val is not None:
         fill_color(objectpath(shape, attr), val)
 
 
-def set_opacity(prop, attr, shape, spec, data: dict):
+def set_opacity(attr, shape, spec, data: dict):
     '''Generator for fill-opacity, stroke-opacity commands'''
     val = expr(spec, data)
     if val is not None:
@@ -338,17 +348,32 @@ def get_elements(children: List[Union[str, HtmlElement]], element_builder):
     Convert head & tail text into specified tag, so that we can just process all text and elements
     as a single list.
     '''
+    elements, open = [], True
     for e in children:
-        if isinstance(e, HtmlElement):
-            tail = e.tail.rstrip() if e.tail else e.tail
-            if e.text or list(e):
-                e.tail = ''
-                yield e if e.tag == element_builder.args[0] else element_builder(e)
-            if tail:
-                yield element_builder(tail)
-        elif isinstance(e, str):
+        if isinstance(e, str):
+            e = re.sub(r'\s+', ' ', e, re.DOTALL)
             if e:
-                yield element_builder(e.rstrip())
+                if open and len(elements):
+                    elements[-1].text += e
+                else:
+                    elements.append(element_builder(e))
+            open = True
+        elif isinstance(e, HtmlElement):
+            if e.tag != element_builder.args[0]:
+                if open and len(elements):
+                    elements[-1].append(e)
+                else:
+                    elements.append(element_builder(e))
+                open = True
+            else:
+                elements.append(e)
+                open = False
+                tail = re.sub(r'\s+', ' ', e.tail, re.DOTALL) if e.tail else ''
+                if tail.strip():
+                    elements.append(element_builder(tail))
+                    open = True
+
+    return elements or [element_builder('')]
 
 
 def baseline(run, val, data):
@@ -368,31 +393,39 @@ def strike(run, val, data):
     run.font._rPr.set('strike', val)
 
 
-para_attr_method = {
-    'align': assign(alignment, 'alignment'),
-    'bold': assign(binary, 'font', 'bold'),
+def get_text_frame(shape):
+    # Get the text frame. Note: don't use .has_text_frame. This function should work for cells too
+    try:
+        return shape.text_frame
+    except AttributeError:
+        raise ValueError('Cannot set text on shape %s that has no text frame' % shape.name)
+
+
+para_methods = {
+    'align': assign(partial(alignment, PP_ALIGN), 'alignment'),
+    'bold': assign(binary, 'font.bold'),
     'color': lambda p, v, d: fill_color(p.font.fill, v),
-    'font-name': assign(str, 'font', 'name'),
-    'font-size': assign(length, 'font', 'size'),
-    'italic': assign(binary, 'font', 'italic'),
+    'font-name': assign(str, 'font.name'),
+    'font-size': assign(length, 'font.size'),
+    'italic': assign(binary, 'font.italic'),
     'level': assign(int, 'level'),
     'line-spacing': assign(length, 'line_spacing'),
     'space-after': assign(length, 'space_after'),
     'space-before': assign(length, 'space_before'),
-    'underline': assign(binary, 'font', 'underline'),
+    'underline': assign(binary, 'font.underline'),
 }
-run_attr_method = {
+run_methods = {
     'baseline': baseline,
-    'bold': assign(binary, 'font', 'bold'),
+    'bold': assign(binary, 'font.bold'),
     'color': lambda r, v, d: fill_color(r.font.fill, v),
-    'font-name': assign(str, 'font', 'name'),
-    'font-size': assign(length, 'font', 'size'),
+    'font-name': assign(str, 'font.name'),
+    'font-size': assign(length, 'font.size'),
     'hover': lambda r, v, d: set_link('text', 'hlinkMouseOver', d['prs'], d['slide'], r, v),
-    'italic': assign(binary, 'font', 'italic'),
+    'italic': assign(binary, 'font.italic'),
     'link': lambda r, v, d: set_link('text', 'hlinkClick', d['prs'], d['slide'], r, v),
     'strike': strike,
     'tooltip': lambda r, v, d: set_tooltip('text', d['prs'], d['slide'], r, v),
-    'underline': assign(binary, 'font', 'underline'),
+    'underline': assign(binary, 'font.underline'),
 }
 
 
@@ -401,11 +434,9 @@ def text(shape, spec, data):
     val = expr(spec, data)
     if val is None:
         return
-    if not shape.has_text_frame:
-        raise ValueError('Cannot add text to shape %s that has no text frame' % shape.name)
+    frame = get_text_frame(shape)
     # Delete all but the first para (required), and its first run (to preserve formatting).
     # All new paras and runs get the same formatting as the first para & run, by default
-    frame = shape.text_frame
     for p in frame.paragraphs[1:]:
         p._p.getparent().remove(p._p)
     for r in frame.paragraphs[0].runs[1:]:
@@ -422,8 +453,8 @@ def text(shape, spec, data):
         p._pPr.attrib.update(para_defaults)
         # Set specified para attributes
         for attr, val in para.attrib.items():
-            if attr in para_attr_method:
-                para_attr_method[attr](p, val, data)
+            if attr in para_methods:
+                para_methods[attr](p, val, data)
         for j, run in enumerate(get_elements([para.text] + list(para), builder.A)):
             # The first run MAY exist, if there was text. Or not. Create if required
             r = p.add_run() if j >= len(p.runs) else p.runs[j]
@@ -432,15 +463,14 @@ def text(shape, spec, data):
             # Set specified run attributes
             r.text = run.text
             for attr, val in run.attrib.items():
-                if attr in run_attr_method:
-                    run_attr_method[attr](r, val, data)
+                if attr in run_methods:
+                    run_methods[attr](r, val, data)
 
 
 def replace(shape, spec, data):
     if not isinstance(spec, dict):
         raise ValueError('replace: needs a dict of {old: new} text, not %r' % spec)
-    if not shape.has_text_frame:
-        raise ValueError('Cannot add text to shape %s that has no text frame' % shape.name)
+    frame = get_text_frame(shape)
 
     def insert_run_after(r, p, original_r):
         new_r = copy.deepcopy(original_r)
@@ -448,7 +478,7 @@ def replace(shape, spec, data):
         return _Run(new_r, p)
 
     spec = {re.compile(old): fragments_fromstring(expr(new, data)) for old, new in spec.items()}
-    for p in shape.text_frame.paragraphs:
+    for p in frame.paragraphs:
         for r in p.runs:
             for old, tree in spec.items():
                 match = old.search(r.text)
@@ -461,27 +491,27 @@ def replace(shape, spec, data):
                         r = insert_run_after(r, p, original_r) if j > 0 else r
                         r.text = run.text
                         for attr, val in run.attrib.items():
-                            if attr in run_attr_method:
-                                run_attr_method[attr](r, val, data)
+                            if attr in run_methods:
+                                run_methods[attr](r, val, data)
                     if suffix:
                         # Ensure suffix has same attrs as original text
                         r = insert_run_after(r, p, original_r)
                         r.text = suffix
 
 
-def set_text(attr, on_para, shape, spec, data):
+def set_text(attr, on_run, shape, spec, data):
     '''Generator for bold, italic, color, and other text commands'''
     val = expr(spec, data)
     if val is not None:
-        if not shape.has_text_frame:
-            raise ValueError('Cannot change text of shape %s that has no text frame' % shape.name)
-        # If on_para=True, set attr on every para, clear attr on every run (e.g. bold, italic)
-        # Else, set attr on every run, clear attr on every para (e.g. color)
-        para_method, run_method = para_attr_method[attr], run_attr_method[attr]
-        for para in shape.text_frame.paragraphs:
-            para_method(para, val if on_para else None, data)
+        frame = get_text_frame(shape)
+        para_method, run_method = para_methods[attr], run_methods[attr]
+        for para in frame.paragraphs:
+            para_method(para, val, data)
             for run in para.runs:
-                run_method(run, None if on_para else val, data)
+                # If on_run=False, clear attr on every run (e.g. bold, italic)
+                # Else, set attr on every run too (e.g. color)
+                # (because PPT doesn't display use color on para.)
+                run_method(run, val if on_run else None, data)
 
 
 # Image commands
@@ -490,14 +520,16 @@ def set_text(attr, on_para, shape, spec, data):
 def image(shape, spec, data: dict):
     val = expr(spec, data)
     if val is not None:
-        rid = shape._pic.blipFill.blip.rEmbed
-        part = shape.part.related_parts[rid]
+        # Load image contents as a bytestring
         if urlparse(val).netloc:
-            part._blob = requests.get(val).content
+            content = requests.get(val).content
         else:
-            part._blob = gramex.cache.open(val, 'bin')
-        # Preserve aspect ratio and width. Adjust height
-        img = Image.open(io.BytesIO(part._blob))
+            content = gramex.cache.open(val, 'bin')
+        # Add the image part
+        image_part, rid = shape.part.get_or_add_image_part(io.BytesIO(content))
+        shape.element.blipFill.blip.rEmbed = rid
+        # Preserve aspect ratio and width. Adjust height via a "cover" algorithm
+        img = Image.open(io.BytesIO(content))
         img_width, img_height = img.size
         shape.height = int(shape.width * img_height / img_width)
 
@@ -516,6 +548,173 @@ def image_height(shape, spec, data: dict):
         shape.width, shape.height = int(val * shape.width / shape.height), val
 
 
+# Table commands
+# ---------------------------------------------------------------------
+def table_align(shape, spec, data: dict):
+    val = expr(spec, data)
+    if val is not None:
+        val = alignment(PP_ALIGN, val)
+        for para in shape.text_frame.paragraphs:
+            para.alignment = val
+
+
+def table_assign(convert, attr, shape, spec, data: dict):
+    val = expr(spec, data)
+    if val is not None:
+        setattr(shape, attr, convert(val))
+
+
+def table_width(table, col_index, spec, data):
+    table._tbl.tblGrid.gridCol_lst[col_index].set('w', str(length(expr(spec, data))))
+
+
+def _resize(elements, n):
+    '''Ensure that a tr, tc or gridCol list has n children by cloning or deleting last element'''
+    for index in range(len(elements), n, -1):
+        elements[index - 1].delete()
+    for index in range(len(elements), n):
+        last_element = copy.deepcopy(elements[-1])
+        # When copying the last element, remove the a:extLst. This may have a rowId and colId
+        # Note: Anand isn't sure if this removes any useful properties. See https://bit.ly/2ZYDw7B
+        for ext_lst in last_element.findall('.//' + qn('a:extLst')):
+            ext_lst.getparent().remove(ext_lst)
+        elements[-1].addnext(last_element)
+
+
+table_cell_commands = {
+    'text': text,
+    'fill': partial(set_color, 'fill'),
+    'fill-opacity': partial(set_opacity, 'fill'),
+    'align': table_align,
+    # TODO: table borders
+    'vertical-align': partial(
+        table_assign, partial(alignment, MSO_VERTICAL_ANCHOR), 'vertical_anchor'),
+    'margin-left': partial(table_assign, length, 'margin_left'),
+    'margin-right': partial(table_assign, length, 'margin_right'),
+    'margin-top': partial(table_assign, length, 'margin_top'),
+    'margin-bottom': partial(table_assign, length, 'margin_bottom'),
+    # TODO: width: column-wise width
+    'bold': partial(set_text, 'bold', False),
+    'color': partial(set_text, 'color', True),     # PPT needs colors on runs too, not only paras
+    'font-name': partial(set_text, 'font-name', False),
+    'font-size': partial(set_text, 'font-size', False),
+    'italic': partial(set_text, 'italic', False),
+    'underline': partial(set_text, 'underline', False),
+}
+table_col_commands = {
+    'width': table_width,
+}
+
+
+def table(shape, spec, data: dict):
+    if not shape.has_table:
+        raise ValueError('Cannot run table commands on shape %s that is not a table' % shape.name)
+    table = shape.table
+
+    # Set or get table first/last row/col attributes
+    header_row = expr(spec.get('header-row', table.first_row), data)
+    table.first_row = bool(header_row)
+    table.last_row = bool(expr(spec.get('total-row', table.last_row), data))
+    table.first_col = bool(expr(spec.get('first-column', table.first_col), data))
+    table.last_col = bool(expr(spec.get('last-column', table.last_col), data))
+
+    # table data must be a DataFrame if specified. Else, create a DataFrame from existing text
+    table_data = expr(spec.get('data', None), data)
+    if table_data is not None and not isinstance(table_data, pd.DataFrame):
+        raise ValueError('data on table %s must be a DataFrame, not %r' % (shape.name, table_data))
+    # Extract data from table text if no data is specified.
+    if table_data is None:
+        table_data = pd.DataFrame([[cell.text for cell in row.cells] for row in table.rows])
+        # If the PPTX table has a header row, set the first row as the DataFrame header too
+        if table.first_row:
+            table_data = table_data.T.set_index([0]).T
+
+    # Adjust PPTX table size to data table size
+    header_offset = 1 if table.first_row else 0
+    _resize(table._tbl.tr_lst, len(table_data) + header_offset)
+    data_cols = len(table_data.columns)
+    _resize(table._tbl.tblGrid.gridCol_lst, data_cols)
+    for row in table.rows:
+        _resize(row._tr.tc_lst, data_cols)
+
+    # Set header row text from header-row or from data column headers
+    if table.first_row:
+        header_columns = table_data.columns
+        if isinstance(header_row, (list, tuple, pd.Index, pd.Series)):
+            header_columns = header_row[:len(table_data.columns)]
+        for j, column in enumerate(header_columns):
+            text(table.cell(0, j), column, {'_expr_mode': False})
+
+    # If `text` is not specified, just use the table value
+    expr_mode = data.get('_expr_mode')
+    spec.setdefault('text', 'cell.val' if expr_mode else {'expr': 'cell.val'})
+
+    # TODO: Handle nans
+    # Apply table commands. (Copy data to avoid modifying original. We'll add data['cell'] later)
+    data = dict(data)
+    columns = table_data.columns.tolist()
+    for key, cmdspec in spec.items():
+        # The command spec can be an expression, or a dict of expressions for each column.
+        # Always convert into a {column: expression}.
+        # But carefully, handling {value: ...} in expr mode and {expr: ...} in literal mode
+        if (not isinstance(cmdspec, dict) or
+                (expr_mode and 'value' in cmdspec) or
+                (not expr_mode and 'expr' in cmdspec)):
+            cmdspec = {column: cmdspec for column in table_data.columns}
+        # Apply commands that run on each cell
+        if key in table_cell_commands:
+            for i, (index, row) in enumerate(table_data.iterrows()):
+                for j, (column, val) in enumerate(row.iteritems()):
+                    data['cell'] = AttrDict(
+                        val=val, column=column, index=index, row=row, data=table_data,
+                        pos=AttrDict(row=i, column=j))
+                    cell = table.cell(i + header_offset, j)
+                    if column in cmdspec:
+                        table_cell_commands[key](cell, cmdspec[column], data)
+            for column in cmdspec:
+                if column not in columns:
+                    app_log.warn('pptgen2: No column: %s in table: %s', column, shape.name)
+        # Apply commands that run on each column
+        elif key in table_col_commands:
+            for column, colspec in cmdspec.items():
+                if column in columns:
+                    col_index = columns.index(column)
+                    data['cell'] = AttrDict(
+                        val=column, column=column, index=None, row=table.columns, data=table_data,
+                        pos=AttrDict(row=-1, column=col_index))
+                    table_col_commands[key](table, col_index, colspec, data)
+                else:
+                    app_log.warn('pptgen2: No column: %s in table: %s', column, shape.name)
+
+
+def chart_data(shape, spec, data: dict):
+    if not hasattr(shape, 'chart'):
+        raise ValueError('Cannot set chart-data on shape: %s' % shape.name)
+    chartdata = pd.read_excel(io.BytesIO(shape.chart.part.chart_workbook.xlsx_part.blob),
+                              index_col=0)
+    val = expr(spec, {'chartdata': chartdata, **data})
+    if val is not None:
+        if not isinstance(val, pd.DataFrame):
+            raise ValueError('chart-data: must be a DataFrame, not %s' % val)
+        chart_tag = shape.chart.plots._plotArea.xCharts[0].tag
+        if chart_tag not in conf['chart-type']:
+            raise ValueError('Shape %s: unsupported chart %s' % (shape.name, chart_tag))
+        chart_type = conf['chart-type'][chart_tag]
+        # Create a new instance of CategoryChartData(), XYChartData() or BubbleChartData()
+        new_chart_data = getattr(pptxchartdata, chart_type + 'ChartData')()
+        new_chart_data.categories = val.index
+        if chart_type == 'Category':
+            for name, col in val.iteritems():
+                new_chart_data.add_series(name, col.values)
+        # TODO: This messes up the resulting Excel sheet, and is not extensible. Rewrite via lxml
+        elif chart_type == 'Xy':
+            for name, col in val.iteritems():
+                series = new_chart_data.add_series(name)
+                for index, v in col.iteritems():
+                    series.add_data_point(index, v)
+        shape.chart.replace_data(new_chart_data)
+
+
 cmdlist = {
     # Basic commands
     'name': name,
@@ -538,10 +737,10 @@ cmdlist = {
     'adjustment3': partial(set_adjustment, 2),
     'adjustment4': partial(set_adjustment, 3),
     # Style
-    'fill': partial(set_color, 'fill', 'fill'),
-    'stroke': partial(set_color, 'stroke', 'line.fill'),
-    'fill-opacity': partial(set_opacity, 'fill-opacity', 'fill'),
-    'stroke-opacity': partial(set_opacity, 'stroke-opacity', 'line.fill'),
+    'fill': partial(set_color, 'fill'),
+    'stroke': partial(set_color, 'line.fill'),
+    'fill-opacity': partial(set_opacity, 'fill'),
+    'stroke-opacity': partial(set_opacity, 'line.fill'),
     'stroke-width': stroke_width,
     # Image
     'image': image,
@@ -557,15 +756,15 @@ cmdlist = {
     # Text
     'replace': replace,
     'text': text,
-    'bold': partial(set_text, 'bold', True),
-    'color': partial(set_text, 'color', False),         # "False" to set color on runs not paras
-    'font-name': partial(set_text, 'font-name', True),
-    'font-size': partial(set_text, 'font-size', True),
-    'italic': partial(set_text, 'italic', True),
-    'underline': partial(set_text, 'underline', True),
+    'bold': partial(set_text, 'bold', False),
+    'color': partial(set_text, 'color', True),  # PPT needs colors on runs too, not only paras
+    'font-name': partial(set_text, 'font-name', False),
+    'font-size': partial(set_text, 'font-size', False),
+    'italic': partial(set_text, 'italic', False),
+    'underline': partial(set_text, 'underline', False),
     # Others
-    # 'chart': chart,
-    # 'table': table,
+    'table': table,
+    'chart-data': chart_data,
     # Custom charts
     # 'sankey': sankey,
     # 'bullet': bullet,
