@@ -1,16 +1,19 @@
 '''Functions to transform data'''
 
-from .transforms import build_transform, condition, flattener, once
+from .transforms import build_transform, build_log_info, condition, flattener, once, handler
 from .badgerfish import badgerfish
-from .template import template, CacheLoader
+from .template import template, sass, scss, CacheLoader
 from .rmarkdown import rmarkdown
 from .auth import ensure_single_session
 from .twitterstream import TwitterStream
 
 __all__ = [
     'build_transform',
+    'build_log_info',
     'badgerfish',
     'template',
+    'sass',
+    'scss',
     'rmarkdown',
     'ensure_single_session',
     'condition',
@@ -18,4 +21,5 @@ __all__ = [
     'once',
     'CacheLoader',
     'TwitterStream',
+    'handler'
 ]
